/* AddMenu.java
 *
 * created: Tue Dec 29 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/AddMenu.java,v 1.28 1999/06/07 15:21:47 kmr Exp $
 */

package diana.components;

import Diana;
import diana.*;
import diana.sequence.*;
import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.Key;
import uk.ac.sanger.pathogens.embl.Location;
import uk.ac.sanger.pathogens.embl.QualifierVector;
import uk.ac.sanger.pathogens.embl.Qualifier;
import uk.ac.sanger.pathogens.embl.QualifierParseException;
import uk.ac.sanger.pathogens.embl.InvalidQualifierException;
import uk.ac.sanger.pathogens.embl.InvalidRelationException;
import uk.ac.sanger.pathogens.embl.InvalidKeyException;
import uk.ac.sanger.pathogens.embl.LocationParseException;

import java.awt.*;
import java.awt.event.*;

/**
 *  A Menu with commands that add new features/entries to an EntryGroup.  This
 *  should be called CreateMenu.
 *
 *  @author Kim Rutherford
 *  @version $Id: AddMenu.java,v 1.28 1999/06/07 15:21:47 kmr Exp $
 **/

public class AddMenu extends SelectionMenu {
  /**
   *  Create a new AddMenu object.
   *  @param frame The Frame that owns this Menu.
   *  @param selection_owner The owner of the selection that the commands in
   *    the menu will operate on.
   *  @param entry_group The EntryGroup object where new features/entries will
   *    be added.
   **/
  public AddMenu (final Frame frame,
                  final SelectionOwner selection_owner,
                  final EntryGroup entry_group) {
    super (frame, "Create", selection_owner);

    this.entry_group = entry_group;

    new_feature_item = new MenuItem ("New Feature");
    new_feature_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        makeNewFeature ();
      }
    });

    add (new_feature_item);

    create_feature_from_range_item =
      new MenuItem ("Create Feature From Base Range");
    create_feature_from_range_item.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        createFeatureFromBaseRange ();
      }
    });

    add (create_feature_from_range_item);

    addSeparator ();

    new_entry_item = new MenuItem ("New Entry");
    new_entry_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        makeNewEntry ();
      }
    });

    add (new_entry_item);

    addSeparator ();

    mark_orfs_item = new MenuItem ("Mark Open Reading Frames");
    mark_orfs_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        markOpenReadingFrames ();
      }
    });

    add (mark_orfs_item);

    mark_orfs_item = new MenuItem ("Mark Open Reading Frames In Range");
    mark_orfs_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        markOpenReadingFramesInRange ();
      }
    });

    add (mark_orfs_item);

    mark_pattern_item = new MenuItem ("Mark From Pattern");
    mark_pattern_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        makeFeaturesFromPattern ();
      }
    });

    add (mark_pattern_item);
  }

  /**
   *  Create a new feature in the default Entry of the entry group.  See
   *  EntryGroup.createFeature () for details.
   **/
  private void makeNewFeature () {
    if (entry_group.size () > 0) {

      if (entry_group.getDefaultEntry () == null) {
        new MessageDialog (getParentFrame (), "There is no default entry");
      } else {

        try {
          final Feature new_feature = entry_group.createFeature ();
          
          new FeatureEdit (new_feature).show ();
        } catch (ReadOnlyException e) {
          new MessageDialog (getParentFrame (), "feature not created: " +
                             "the default entry is read only");
        }
      }
    } else {
      new MessageDialog (getParentFrame (),
                         "Cannot make a feature without an existing entry");
    }
  }

  /**
   *  Create a new Entry in the first Entry of the entry group.
   **/
  private void makeNewEntry () {
    entry_group.createEntry ();
  }

  /**
   *  Create a new Feature in entry_group from the selected range of bases.
   **/
  private void createFeatureFromBaseRange () {
    if (!checkForSelectionRange ()) {
      return;
    }

    final MarkerRange range =
      getSelectionOwner ().getSelection ().getMarkerRange ();
    final Entry default_entry = entry_group.getDefaultEntry ();

    if (default_entry == null) {
      new MessageDialog (getParentFrame (), "There is no default entry");
      return;
    }
    
    try {
      final uk.ac.sanger.pathogens.embl.Location new_location =
        range.createLocation ();

      final Feature new_feature =
        default_entry.createFeature (Key.CDS, new_location);
      getSelectionOwner ().getSelection ().setMarkerRange (null);
      getSelectionOwner ().getSelection ().set (new_feature);

      new FeatureEdit (new_feature).show ();
    } catch (LocationParseException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    } catch (ReadOnlyException e) {
      new MessageDialog (getParentFrame (), "feature not created: " +
                         "the default entry is read only");
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  Create a new Feature for each open reading frame.  The minimum size of
   *  the ORFS is specified in the options file.
   **/
  private void markOpenReadingFrames () {
    try {
      final int minimum_orf_size = Diana.getOptions ().getMinimumORFSize ();

      final Entry new_entry =
        entry_group.createEntry ("ORFS_" + minimum_orf_size + '+');

      final int sequence_length = entry_group.getSequenceLength ();

      final Strand forward_strand =
        entry_group.getBases ().getForwardStrand ();

      final MarkerRange forward_sequence_range =
        forward_strand.makeMarkerRangeFromPositions (1, sequence_length);

      markOpenReadingFrames (new_entry, forward_sequence_range);

      final Strand backward_strand =
        entry_group.getBases ().getBackwardStrand ();

      final MarkerRange backward_sequence_range =
        backward_strand.makeMarkerRangeFromPositions (1, sequence_length);

      markOpenReadingFrames (new_entry, backward_sequence_range);
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected OutOfRangeException");
    }
  }

  /**
   *  Create a new Feature for each open reading frame.  The minimum size of
   *  the ORFS is specified in the options file.
   **/
  private void markOpenReadingFramesInRange () {
    if (!checkForSelectionRange ()) {
      return;
    }

    final MarkerRange selection_range =
      getSelectionOwner ().getSelection ().getMarkerRange ();

    final int minimum_orf_size = Diana.getOptions ().getMinimumORFSize ();

    final Entry new_entry =
      entry_group.createEntry ("ORFS_" + minimum_orf_size + '+');

    final int sequence_length = entry_group.getSequenceLength ();

    markOpenReadingFrames (new_entry, selection_range);
  }

  /**
   *  Create a new Feature in the given Entry for each open reading frame that
   *  overlaps the given range.  The minimum size of the ORFS is specified in
   *  the options file.
   *  @param entry The new features are created in this entry.
   **/
  private void markOpenReadingFrames (final Entry entry,
                                      final MarkerRange search_range) {
    final int minimum_orf_size = Diana.getOptions ().getMinimumORFSize ();

    final MarkerRange [] forward_orf_ranges =
      Strand.getOpenReadingFrameRanges (search_range, minimum_orf_size);

    final Key key;

    try {
      key = new Key ("CDS");
    } catch (InvalidKeyException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }

    for (int i = 0 ; i < forward_orf_ranges.length ; ++i) {
      final MarkerRange this_range = forward_orf_ranges[i];
      try {
        makeFeatureFromMarkerRange (entry, this_range, key);
      }  catch (ReadOnlyException e) {
        new MessageDialog (getParentFrame (), "cannot continue: " +
                           "the default entry is read only");
        return;
      }
    }
  }

  /**
   *  Make a new Feature from the given MarkerRange in the given Entry.  The
   *  new feature will be given the key 'CDS' and it's location will match the
   *  MarkerRange.
   *  @param entry The new feature is created in this entry.
   *  @param range The location of the new feature.
   *  @param key The key give the new feature
   **/
  private Feature makeFeatureFromMarkerRange (final Entry entry,
                                              final MarkerRange range,
                                              final Key key)
      throws ReadOnlyException {
    try {
      final Location new_location = range.createLocation ();

      final QualifierVector qualifiers = new QualifierVector ();

      try {
        qualifiers.setQualifier (new Qualifier ("note", "none"));
        qualifiers.setQualifier (new Qualifier ("label", "*"));
        qualifiers.setQualifier (new Qualifier ("gene", "none"));
      } catch (QualifierParseException e) {
        throw new Error ("internal error - unexpected exception: " + e);
      } catch (InvalidQualifierException e) {
        throw new Error ("internal error - unexpected exception: " + e);
      }
      
      final Feature new_feature =
        entry.createFeature (key, new_location, qualifiers);

      return new_feature;
    } catch (LocationParseException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    } catch (InvalidRelationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  This method will ask the user for a BasePattern (using a TextRequester
   *  component) then search the sequence for the given pattern and make a new
   *  feature from each match.  The new features will created in an Entry
   *  called "matches: <pattern>".
   **/
  private void makeFeaturesFromPattern () {
    final TextRequester text_requester =
      new TextRequester ("create features from this pattern:", 18, "");

    text_requester.addTextRequesterListener (new TextRequesterListener () {
      public void actionPerformed (final TextRequesterEvent event) {
        final String pattern_string = event.getRequesterText ().trim ();

        try {
          if (pattern_string.length () == 0) {
            new MessageDialog (getParentFrame (), "the pattern is too short");
            return;
          }

          final BasePattern pattern = new BasePattern (pattern_string);

          makeFeaturesFromPattern (pattern);
        } catch (BasePatternFormatException e) {
          new MessageDialog (getParentFrame (),
                             "Illegal base pattern: " +
                             pattern_string);
        }
      }
    });

    text_requester.show ();
  }

  /**
   *  Search the sequence for the given pattern and make a new feature from
   *  each match.  The new features will created in an Entry called "matches:
   *  <pattern>".
   **/
  private void makeFeaturesFromPattern (final BasePattern pattern) {
    final MarkerRangeVector matches =
      pattern.findMatches (entry_group.getBases (),
                           null,        // search from start
                           entry_group.getSequenceLength ());

    if (matches.size () == 0) {
      new MessageDialog (getParentFrame (),
                         "no matches found for: " + pattern);
      return;
    }

    final int TOO_MANY_MATCHES = 100;

    if (matches.size () > TOO_MANY_MATCHES) {
      final YesNoDialog dialog =
        new YesNoDialog (getParentFrame (),
                         matches.size () + " matches, continue?");

      if (dialog.getResult ()) {
        // yes - continue
      } else {
        // no
        return;
      }
    }

    final Entry new_entry = entry_group.createEntry ("matches: " + pattern);

    final Key key;

    try {
      key = new Key ("misc_feature");
    } catch (InvalidKeyException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }

    for (int i = 0 ; i < matches.size () ; ++i) {
      try {
        final Feature new_feature =
          makeFeatureFromMarkerRange (new_entry, matches.elementAt (i), key);
        new_feature.setQualifier (new Qualifier ("note", pattern.toString ()));
      } catch (QualifierParseException e) {
        throw new Error ("internal error - unexpected exception: " + e);
      } catch (InvalidQualifierException e) {
        throw new Error ("internal error - unexpected exception: " + e);
      } catch (InvalidRelationException e) {
        throw new Error ("internal error - unexpected exception: " + e);
      } catch (ReadOnlyException e) {
        new MessageDialog (getParentFrame (), "cannot continue: " +
                           "the default entry is read only");
        return;
      }
    }
  }

  /**
   *  The EntryGroup object that was passed to the constructor.
   **/
  final EntryGroup entry_group;

  private MenuItem new_feature_item = null;
  private MenuItem new_entry_item = null;
  private MenuItem create_feature_from_range_item = null;
  private MenuItem mark_orfs_item = null;
  private MenuItem mark_pattern_item = null;
}


